type 'a t = Empty | Node of ('a t * 'a * 'a t);;

let empty = Empty;;
let singleton x = Node (Empty, x, Empty);;
let rec insert x = function
 | Empty -> singleton x
 | Node (_, v, _) as n when x = v -> n
 | Node (lt, v, rt) when x < v -> Node (insert x lt, v, rt)
 | Node (lt, v, rt) when x > v -> Node (lt, v, insert x rt)
;;
let rec union xt = function
 | Empty -> xt
 | Node (lt, v, rt) -> union (union (insert v xt) lt) rt
;;
let rec remove_min = function
 | Empty -> (None, Empty)
 | Node (Empty, v, rt) -> (Some v, rt)
 | Node (lt, v, rt) -> let (m, lt) = remove_min lt in (m, Node (lt, v, rt))
;;
let rec remove_max = function
 | Empty -> (None, Empty)
 | Node (lt, v, Empty) -> (Some v, lt)
 | Node (lt, v, rt) -> let (m, rt) = remove_max rt in (m, Node (lt, v, rt))
;;
let rec remove x = function
 | Empty -> Empty
 | Node (lt, v, rt) when x = v ->
  begin match remove_max lt with
   | (None, _) -> rt
   | (Some m, lt) -> Node (lt, m, rt)
  end
 | Node (lt, v, rt) when x < v -> Node (remove x lt, v, rt)
 | Node (lt, v, rt) when x > v -> Node (lt, v, remove x rt)
;;
let rec diff xt = function
 | Empty -> xt
 | Node (lt, v, rt) -> diff (remove v xt) (union lt rt)
;;
let rec of_list = function 
 | [] -> Empty
 | x :: xs -> insert x (of_list xs)
;;
let rec to_list xt = match remove_min xt with
 | (None, _) -> []
 | (Some m, xt) -> m :: to_list xt
;;
let rec mem x = function
 | Empty -> false
 | Node (_, v, _) when x = v -> true
 | Node (lt, v, _) when x < v -> mem x lt
 | Node (_, v, rt) when x > v -> mem x rt
;;
let is_empty xt = xt = Empty;;
let rec min = function
 | Empty -> failwith "Set.min: empty set"
 | Node (Empty, v, _) -> v
 | Node (lt, _, _) -> min lt
;;
let choose xt = min xt;;
let unwrap = function
 | None -> failwith "St.unwrap: empty option"
 | Some x -> x
;;
let rec fold f b ts =
 if ts = Empty then 
  b
 else
  let (m, ts) = remove_min ts in
  f (unwrap m) (fold f b ts)
;;
let filter p ts = fold (fun x d -> if p x then insert x d else d) empty ts;;
let is_singleton = function
 | Node (Empty, _, Empty) -> true
 | _ -> false
;;
let print_st print_element fmt ts =
 let rec print_st fmt ts =
  if ts = Empty then
   ()
  else if is_singleton ts then
   Format.fprintf fmt "%a" print_element (min ts) 
  else
   let (m, rest) = remove_min ts in
   Format.fprintf fmt "%a, %a" print_element (unwrap m) print_st rest
 in
 Format.fprintf fmt "{%a}" print_st ts
;;
